--- FSM and Top Module for the DHBW MCT controller (VHDL)

use work.MCT_Pack_uP.all;
library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;

entity FSM_Top_Module is
	 port ( Clk, RS: in std_logic);  
end FSM_Top_Module;

architecture RTL of FSM_Top_Module  is

component P_ROM is           
	port ( Clk, Pipe_EN  : in std_logic; -- Clock, Pipeline enable
		     PADDR : in  PA_Type;      
		     PDOUT : out P_Type);    	
end component;

component D_RAM is           
	port ( Clk, DS_EN, RnW, SEL_PC_Akku  : in std_logic; -- Clock, Pipeline enable
		     RWADDR : in  DA_Type;
		     DATIN, StackO : in D_Type;      
		     DTOUT, StackI : out D_Type);    	
end component;

component IR_DEC is
	 port ( Clk: in std_logic; RS : in std_logic; DIN : in P_Type;
	        PC_EN, Pipe_EN, PC_LOADA, PC_LOADR, PC_LOADS, 
	        SP_INC, SP_DEC : in std_logic;
	        A : out DA_Type; -- addresses of operands
	        PSR_Mask : out std_logic_vector(4 downto 0);
	        PC : out PA_Type; OPCODE : out OPTYPE; K8, 
	        StackO : out D_Type; StackI : in D_Type);   
end component;

component ALU_PSR_Akku is
	 port ( OPCODE : in OPTYPE; K, DS_DO : in D_Type; Clk, 
	        ACC_EN : in std_logic; DS_DI  : out D_Type; 
	        CFlag, ZFlag, NFlag: out std_logic); 
end component;

  
-- Top_Module internal signals

type state_type is (Z1, Z2, Z3); 
-- Z1: operation executes in one clock cycle, 
-- Z2: operation executes in two clock cycles,
-- Z3: operation executes in two clock cycles  

signal next_state, current_state: state_type :=Z1;

-- to interconnect modules incl. controls
signal T_OPCODE : OPTYPE;
signal T_PC : PA_TYPE;
signal T_A : DA_TYPE;
signal T_DIN : P_TYPE;
signal T_DI, T_DO, T_K8, T_StackO, T_StackI : D_Type;
signal T_Zero, T_Carry, T_Neg : std_logic;
signal T_PSR : std_logic_vector (4 downto 0);
signal T_Pipe_EN, T_PC_EN, T_DS_RnW : std_logic :='1';
signal T_DS_EN, T_PC_LOADA, T_PC_LOADR, T_PC_LOADS, T_ACC_EN, 
T_SP_INC, T_SP_DEC, T_SEL_PC_Akku : std_logic :='0';

begin
  
-- instantiate and connect modules to Top Module
PMem: P_ROM port map (Clk => Clk, Pipe_EN=> T_Pipe_EN, PADDR => T_PC, PDOUT => T_DIN);
 
-- data memory  
DMem: D_RAM port map (Clk => Clk, DS_EN=> T_DS_EN, RnW => T_DS_RnW, 
                      SEL_PC_Akku => T_SEL_PC_Akku, RWADDR => T_A, 
                      DATIN => T_DI, StackO => T_StackO, 
                      DTOUT => T_DO, StackI => T_StackI);                  
 
IR_Decoder: IR_DEC port map (Clk => Clk, RS => RS, DIN => T_DIN, 
                   Pipe_EN => T_Pipe_EN, PC_EN => T_PC_EN, PC_LOADA => T_PC_LOADA, 
                   PC_LOADR => T_PC_LOADR, PC_LOADS => T_PC_LOADS, 
                   SP_INC => T_SP_INC, SP_DEC => T_SP_DEC,
               	   A => T_A, PSR_Mask => T_PSR, PC => T_PC, OPCODE => T_OPCODE, 
               	   K8 => T_K8, StackO => T_StackO, StackI => T_StackI);         

Alu : ALU_PSR_Akku port map (OPCODE => T_OPCODE, K => T_K8, DS_DO => T_DO, 
                   Clk => Clk, ACC_EN => T_ACC_EN, DS_DI => T_DI, 
                   ZFlag => T_Zero, CFlag => T_Carry, NFlag => T_Neg); 

-- run processes
update_state_register: process(Clk, RS) 
  begin
       if (RS='1') then
         current_state <= Z1;
       elsif rising_edge(Clk) then
         current_state <= next_state;
       end if;

end process update_state_register;

-- state transitions & actions (single logic)
logic_next_state_and_actions: process(current_state, T_OPCODE)
  
  begin
    
    case current_state is
    
    when Z1 => -- one clock cycle per operation
          
           case T_OPCODE is
           when NOP => T_PC_EN<='1'; T_Pipe_EN<='1'; T_DS_EN<='0'; 
                       T_PC_LOADA<='0'; T_PC_LOADR<='0'; T_ACC_EN<='0';
                       T_SP_DEC<='0';   T_SP_INC<='0'; T_PC_LOADS<='0';
                       T_SEL_PC_Akku<='0';
             
           when LSL | LSR | LDI => T_Pipe_EN<='1'; T_PC_EN<='1'; 
                                   T_ACC_EN<='1'; T_DS_EN<='0';  
                                   T_SP_DEC<='0'; T_SP_INC<='0'; T_PC_LOADS<='0';
                                   T_SEL_PC_Akku<='0';
           
           when STR => T_Pipe_EN<='1'; T_PC_EN<='1'; 
                       T_DS_EN<='1'; T_DS_RnW<='0'; T_ACC_EN<='0';
                       T_SP_DEC<='0'; T_SP_INC<='0'; T_PC_LOADS<='0';
                       T_SEL_PC_Akku<='0';
           
           when ADD | SUB | ANDA | EOR | ORA | LDA  => 
                        T_Pipe_EN<='0'; T_ACC_EN<='0'; 
                        T_DS_EN<='1'; T_DS_RnW<='1'; T_SEL_PC_Akku<='0';
                        T_SP_DEC<='0';   T_SP_INC<='0'; T_PC_LOADS<='0';
                        next_state<=Z2;  -- 2nd cycle            
                           
           when JMP =>  T_Pipe_EN<='0'; T_PC_EN<='1'; T_ACC_EN<='0'; 
                        T_PC_LOADA<='1'; T_SEL_PC_Akku<='0';
                        T_SP_DEC<='0';   T_SP_INC<='0'; T_PC_LOADS<='0';
                        next_state<=Z2;  -- 2nd cycle   
           
           when PUSH => T_Pipe_EN<='1'; T_PC_EN<='1'; T_DS_EN <='1';
                        T_PC_LOADA<='0'; T_PC_LOADR<='0'; T_PC_LOADS<='0';
                        T_SP_INC<='1';   T_SP_DEC<='0'; 
                        -- enable stack memory for write operation
                        T_SEL_PC_Akku<='0';  T_DS_RnW<='0';
                        
           when POP =>  T_Pipe_EN<='0';  T_PC_EN<='1'; T_ACC_EN<='0'; 
                        T_PC_LOADA<='0'; T_PC_LOADR<='0'; T_PC_LOADS<='0';
                        T_DS_EN <='1';  T_DS_RnW<='1'; T_SEL_PC_Akku<='0';
                        T_SP_DEC<='1';   T_SP_INC<='0'; 
                        next_state<=Z2;  -- 2nd cycle  
                        
           when CALL =>  T_Pipe_EN<='0'; T_PC_EN<='1'; T_ACC_EN<='0'; 
                        T_PC_LOADA<='1'; T_PC_LOADR<='0'; 
                        T_DS_RnW<='0'; T_DS_EN <='1'; T_SEL_PC_Akku<='1';
                        T_PC_LOADS<='0'; T_SP_INC<='1';  T_SP_DEC<='0';
                        next_state<=Z2;  -- 2nd cycle   
                        
           when RET =>  T_Pipe_EN<='0';  T_PC_EN<='1'; T_ACC_EN<='0'; 
                        T_PC_LOADA<='0'; T_PC_LOADR<='0'; T_ACC_EN<='0'; 
                        T_DS_EN <='1';   T_DS_RnW<='1'; T_SEL_PC_Akku<='1';
                        T_PC_LOADS<='0'; T_SP_DEC<='1'; T_SP_INC<='0'; 
                        next_state<=Z2;  -- 2nd cycle    
           
           when BRBC => 
                        case T_PSR is
                        when "10000" => if (T_Zero = '0')  then  
                                             T_Pipe_EN<='0'; T_PC_LOADR<='1'; 
                                             T_PC_LOADS<='0'; T_ACC_EN<='0';  
                                             next_state<=Z2;
                                        elsif (T_Zero ='1') then
                                             T_Pipe_EN<='1'; T_PC_LOADR<='0'; 
                                             T_PC_LOADS<='0'; T_ACC_EN<='0'; 
                                        end if;
                        when "01000" => if (T_Carry = '0') then  
                                             T_Pipe_EN<='0'; T_PC_LOADR<='1';
                                             T_PC_LOADS<='0'; T_ACC_EN<='0'; 
                                             next_state<=Z2;
                                        elsif (T_Zero ='1') then
                                             T_Pipe_EN<='1'; T_PC_LOADR<='0';
                                             T_PC_LOADS<='0'; T_ACC_EN<='0'; 
                                        end if;
                        when "00001" => if (T_Neg = '0') then  
                                             T_Pipe_EN<='0'; T_PC_LOADR<='1';
                                             T_PC_LOADS<='0'; T_ACC_EN<='0'; 
                                             next_state<=Z2;
                                        elsif (T_Neg ='1') then
                                             T_Pipe_EN<='1'; T_PC_LOADR<='0'; 
                                             T_PC_LOADS<='0'; T_ACC_EN<='0'; 
                                        end if;
                        when others => null;
                        end case;                 
           
           when BRBS =>  
                        case T_PSR is
                        when "10000" => if (T_Zero = '1')  then  
                                             T_Pipe_EN<='0'; T_PC_LOADR<='1';
                                             T_PC_LOADS<='0'; T_ACC_EN<='0'; 
                                             next_state<=Z2;
                                        elsif (T_Zero ='0') then
                                             T_Pipe_EN<='1'; T_PC_LOADR<='0';
                                             T_PC_LOADS<='0'; T_ACC_EN<='0'; 
                                        end if;

                        when "01000" => if (T_Carry = '1') then  
                                             T_Pipe_EN<='0'; T_PC_LOADR<='1';
                                             T_PC_LOADS<='0'; T_ACC_EN<='0'; 
                                             next_state<=Z2; 
                                        elsif (T_Zero ='0') then
                                             T_Pipe_EN<='1'; T_PC_LOADR<='0'; 
                                             T_PC_LOADS<='0'; T_ACC_EN<='0'; 
                                        end if; 
                        when "00001" => if (T_Neg = '1') then  
                                             T_Pipe_EN<='0'; T_PC_LOADR<='1';
                                             T_PC_LOADS<='0'; T_ACC_EN<='0'; 
                                             next_state<=Z2; 
                                        elsif (T_Neg ='0') then
                                             T_Pipe_EN<='1'; T_PC_LOADR<='0';
                                             T_PC_LOADS<='0'; T_ACC_EN<='0'; 
                                        end if;                                  
                        when others => null;
                        end case;  
           
           when others  => null;
           
           end case; -- opcode one cycle
         
      when Z2 => -- second clock cycle per operation
        
        case T_OPCODE is  

           when BRBS | BRBC| JMP | CALL => -- re-animate pipeline
                T_PC_EN<='1'; T_Pipe_EN<='1'; T_PC_LOADA<='0'; 
                T_PC_LOADR<='0'; T_PC_LOADS<='0';
                T_SP_INC <= '0'; T_SP_DEC <= '0'; T_DS_EN <='0';
                next_state<=Z1;

          when RET => -- keep pipeline disabled 
                         T_SP_DEC<='0'; T_SP_INC<='0';
                      -- keep control signals from state Z1
                -- load PC from stack
                T_PC_LOADS<='1';
                next_state<=Z3; -- 3rd cycle
                
          when POP => T_ACC_EN<='1'; T_DS_EN<='0'; -- enable akku
                -- re-animate pipeline
                T_PC_EN<='1'; T_Pipe_EN<='1'; 
                T_PC_LOADR<='0'; T_PC_LOADS<='0';
                T_SP_INC <= '0'; T_SP_DEC <= '0';
                next_state<=Z1;         

                                              
           when ADD | SUB | ANDA | EOR | ORA | LDA  => 
                T_ACC_EN<='1'; T_DS_EN<='0'; -- enable akku for result
                T_Pipe_EN<='1';     -- re-animate pipeline
                next_state<=Z1;  
                        
           when others  => null;
        end case; -- opcode two cycles 
        
      when Z3 => -- third clock cycle per operation
        
        case T_OPCODE is  

          when RET => -- re-animate pipeline
                T_PC_EN<='1'; T_Pipe_EN<='1'; T_PC_LOADS<='0';
                next_state<=Z1; 
                        
           when others  => null;
        end case; -- opcode three cycles 


	    when others => null;
	  end case; -- states

end process logic_next_state_and_actions;

end RTL;

